
#if defined (HAVE_CONFIG_H)
#  include "config.h"
#endif

#include <signal.h>

/* check for standard C headers */
#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <pthread.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <libq.h>
#include <portaudio.h>

#define sys_to_utf8(s) to_utf8(s, NULL)
#define utf8_to_sys(s) from_utf8(s, NULL)

MODULE(audio)

/* The timing infos of the current PortAudio v19 in CVS seem to be broken,
   hence we do the necessary bookkeeping ourselves. This can be commented out
   once PortAudio's own timing facilities work ok. */
#define TIMER_KLUDGE

#define MODE_READ 0x01
#define MODE_WRITE 0x02
#define MODE_RDWR 0x03

static int init_ok = 0;

/* handle SIGINT and SIGTERM */

#ifdef _MSC_VER
#define SIGHANDLER_RETURN(status) return
#else
#if RETSIGTYPE == void
#define SIGHANDLER_RETURN(status) return
#else
#define SIGHANDLER_RETURN(status) return status
#endif
#endif

#ifdef MUST_REINSTALL_SIGHANDLERS
#define SIGHANDLER_RESTORE(sig,handler) syssignal(sig,handler)
#else
#define SIGHANDLER_RESTORE(sig,handler) /* nop */
#endif

typedef RETSIGTYPE (*sighandler_t)(int);

static sighandler_t syssignal(sig, handler)
     int sig;
     sighandler_t handler;
{
#ifdef HAVE_POSIX_SIGNALS
  struct sigaction new_action, old_action;
  new_action.sa_handler = handler;
  sigemptyset(&new_action.sa_mask);
  sigemptyset(&old_action.sa_mask);
  new_action.sa_flags = 0;
  sigaction(sig, &new_action, &old_action);
  return old_action.sa_handler;
#else
  return signal(sig, handler);
#endif
}

static volatile int brkflag = 0;
static sighandler_t int_handler = NULL, term_handler = NULL,
  hup_handler = NULL;

static RETSIGTYPE break_handler(int sig)
{
  if (sig == SIGINT && int_handler) int_handler(sig);
  if (sig == SIGTERM && term_handler) term_handler(sig);
#ifdef SIGHUP
  if (sig == SIGHUP && hup_handler) hup_handler(sig);
#endif
  SIGHANDLER_RESTORE(sig, break_handler);
  brkflag = 1;
  SIGHANDLER_RETURN(0);
}

FUNCTION(audio,audio_vars,argc,argv)
{
  if (argc != 0) return __FAIL;
  return mktuplel
    (13,
     mkint(paFloat32),
     mkint(paInt32),
     mkint(paInt24),
     mkint(paInt16),
     mkint(paInt8),
     mkint(paUInt8),
     mkint(paCustomFormat),
     mkint(MODE_READ),
     mkint(MODE_WRITE),
     mkint(MODE_RDWR),
     mkint(init_ok?Pa_GetDefaultInputDevice():paNoDevice),
     mkint(init_ok?Pa_GetDefaultOutputDevice():paNoDevice),
     mkint(init_ok?Pa_GetDefaultHostApi():(-1)));
}

FUNCTION(audio,audio_drivers,argc,argv)
{
  int count, api, i, n;
  expr *xlist;
  if (!init_ok || argc != 0) return __FAIL;
  count = Pa_GetHostApiCount();
  if (count <= 0) return mknil;
  if (!(xlist = malloc(count*sizeof(expr))))
    return __ERROR;
  for (api = 0; api < count; api++) {
    expr devs, *xv;
    const PaHostApiInfo *info = Pa_GetHostApiInfo(api);
    if (info) {
      int i, n = info->deviceCount;
      if (n == 0)
	devs = mknil;
      else {
	if (!(xv = malloc(n*sizeof(expr)))) {
	  while (api > 0) dispose(xlist[--api]);
	  free(xlist);
	  return __ERROR;
	}
	for (i = 0; i < n; i++)
	  xv[i] = mkint(Pa_HostApiDeviceIndexToDeviceIndex(api, i));
	devs = mklistv(n, xv);
      }
      xlist[api] = mktuplel(5, mkstr(sys_to_utf8(info->name)),
			    mkint(info->type), devs,
			    mkint(info->defaultInputDevice),
			    mkint(info->defaultOutputDevice));
    } else
      xlist[api] = mkvoid;
  }
  return mklistv(count, xlist);
}

FUNCTION(audio,audio_devices,argc,argv)
{
  int count, dev, i, n;
  expr *xlist;
  if (!init_ok || argc != 0) return __FAIL;
  count = Pa_GetDeviceCount();
  if (count <= 0) return mknil;
  if (!(xlist = malloc(count*sizeof(expr))))
    return __ERROR;
  for (dev = 0; dev < count; dev++) {
    const PaDeviceInfo *info = Pa_GetDeviceInfo(dev);
    if (info)
      xlist[dev] = mktuplel(5, mkstr(sys_to_utf8(info->name)),
			    mkint(info->hostApi),
			    mkint(info->maxInputChannels),
			    mkint(info->maxOutputChannels),
			    mkfloat(info->defaultSampleRate));
    else
      xlist[dev] = mkvoid;
  }
  return mklistv(count, xlist);
}

/*
 * This code pilfered from PortAudio. We include our own static version of the
 * ringbuffer utilities here so that the critical routines can be inlined and
 * to prevent conflicts with recent PortAudio 19 versions which expose the
 * RingBuffer interface.
 *
 * ringbuffer.c
 * Ring Buffer utility..
 *
 * Author: Phil Burk, http://www.softsynth.com
 *
 * This program uses the PortAudio Portable Audio Library.
 * For more information see: http://www.audiomulch.com/portaudio/
 * Copyright (c) 1999-2000 Ross Bencina and Phil Burk
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * Any person wishing to distribute modifications to the Software is
 * requested to send the modifications to the original developer so that
 * they can be incorporated into the canonical version.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

typedef struct
{
    long   bufferSize; /* Number of bytes in FIFO. Power of 2. Set by RingBuffer_Init. */
/* These are declared volatile because they are written by a different thread than the reader. */
    volatile long   writeIndex; /* Index of next writable byte. Set by RingBuffer_AdvanceWriteIndex. */
    volatile long   readIndex;  /* Index of next readable byte. Set by RingBuffer_AdvanceReadIndex. */
    long   bigMask;    /* Used for wrapping indices with extra bit to distinguish full/empty. */
    long   smallMask;  /* Used for fitting indices to buffer. */
    char *buffer;
}
MyRingBuffer;

/***************************************************************************
 * Initialize FIFO.
 * numBytes must be power of 2, returns -1 if not.
 */

static void MyRingBuffer_Flush( MyRingBuffer *rbuf );

static long MyRingBuffer_Init( MyRingBuffer *rbuf, long numBytes, void *dataPtr )
{
    if( ((numBytes-1) & numBytes) != 0) return -1; /* Not Power of two. */
    rbuf->bufferSize = numBytes;
    rbuf->buffer = (char *)dataPtr;
    MyRingBuffer_Flush( rbuf );
    rbuf->bigMask = (numBytes*2)-1;
    rbuf->smallMask = (numBytes)-1;
    return 0;
}
/***************************************************************************
** Return number of bytes available for reading. */
static inline long MyRingBuffer_GetReadAvailable( MyRingBuffer *rbuf )
{
    return ( (rbuf->writeIndex - rbuf->readIndex) & rbuf->bigMask );
}
/***************************************************************************
** Return number of bytes available for writing. */
static inline long MyRingBuffer_GetWriteAvailable( MyRingBuffer *rbuf )
{
    return ( rbuf->bufferSize - MyRingBuffer_GetReadAvailable(rbuf));
}

/***************************************************************************
** Clear buffer. Should only be called when buffer is NOT being read. */
static void MyRingBuffer_Flush( MyRingBuffer *rbuf )
{
    rbuf->writeIndex = rbuf->readIndex = 0;
}

/***************************************************************************
** Get address of region(s) to which we can write data.
** If the region is contiguous, size2 will be zero.
** If non-contiguous, size2 will be the size of second region.
** Returns room available to be written or numBytes, whichever is smaller.
*/
static inline long MyRingBuffer_GetWriteRegions( MyRingBuffer *rbuf, long numBytes,
						 void **dataPtr1, long *sizePtr1,
						 void **dataPtr2, long *sizePtr2 )
{
    long   index;
    long   available = MyRingBuffer_GetWriteAvailable( rbuf );
    if( numBytes > available ) numBytes = available;
    /* Check to see if write is not contiguous. */
    index = rbuf->writeIndex & rbuf->smallMask;
    if( (index + numBytes) > rbuf->bufferSize )
    {
        /* Write data in two blocks that wrap the buffer. */
        long   firstHalf = rbuf->bufferSize - index;
        *dataPtr1 = &rbuf->buffer[index];
        *sizePtr1 = firstHalf;
        *dataPtr2 = &rbuf->buffer[0];
        *sizePtr2 = numBytes - firstHalf;
    }
    else
    {
        *dataPtr1 = &rbuf->buffer[index];
        *sizePtr1 = numBytes;
        *dataPtr2 = NULL;
        *sizePtr2 = 0;
    }
    return numBytes;
}


/***************************************************************************
*/
static inline long MyRingBuffer_AdvanceWriteIndex( MyRingBuffer *rbuf, long numBytes )
{
    return rbuf->writeIndex = (rbuf->writeIndex + numBytes) & rbuf->bigMask;
}

/***************************************************************************
** Get address of region(s) from which we can read data.
** If the region is contiguous, size2 will be zero.
** If non-contiguous, size2 will be the size of second region.
** Returns room available to be written or numBytes, whichever is smaller.
*/
static inline long MyRingBuffer_GetReadRegions( MyRingBuffer *rbuf, long numBytes,
						void **dataPtr1, long *sizePtr1,
						void **dataPtr2, long *sizePtr2 )
{
    long   index;
    long   available = MyRingBuffer_GetReadAvailable( rbuf );
    if( numBytes > available ) numBytes = available;
    /* Check to see if read is not contiguous. */
    index = rbuf->readIndex & rbuf->smallMask;
    if( (index + numBytes) > rbuf->bufferSize )
    {
        /* Write data in two blocks that wrap the buffer. */
        long firstHalf = rbuf->bufferSize - index;
        *dataPtr1 = &rbuf->buffer[index];
        *sizePtr1 = firstHalf;
        *dataPtr2 = &rbuf->buffer[0];
        *sizePtr2 = numBytes - firstHalf;
    }
    else
    {
        *dataPtr1 = &rbuf->buffer[index];
        *sizePtr1 = numBytes;
        *dataPtr2 = NULL;
        *sizePtr2 = 0;
    }
    return numBytes;
}
/***************************************************************************
*/
static inline long MyRingBuffer_AdvanceReadIndex( MyRingBuffer *rbuf, long numBytes )
{
    return rbuf->readIndex = (rbuf->readIndex + numBytes) & rbuf->bigMask;
}

/***************************************************************************
** Return bytes written. */
static inline long MyRingBuffer_Write( MyRingBuffer *rbuf, const void *data, long numBytes )
{
    long size1, size2, numWritten;
    void *data1, *data2;
    numWritten = MyRingBuffer_GetWriteRegions( rbuf, numBytes, &data1, &size1, &data2, &size2 );
    if( size2 > 0 )
    {

        memcpy( data1, data, size1 );
        data = ((char *)data) + size1;
        memcpy( data2, data, size2 );
    }
    else
    {
        memcpy( data1, data, size1 );
    }
    MyRingBuffer_AdvanceWriteIndex( rbuf, numWritten );
    return numWritten;
}

/***************************************************************************
** Return bytes read. */
static inline long MyRingBuffer_Read( MyRingBuffer *rbuf, void *data, long numBytes )
{
    long size1, size2, numRead;
    void *data1, *data2;
    numRead = MyRingBuffer_GetReadRegions( rbuf, numBytes, &data1, &size1, &data2, &size2 );
    if( size2 > 0 )
    {
        memcpy( data, data1, size1 );
        data = ((char *)data) + size1;
        memcpy( data, data2, size2 );
    }
    else
    {
        memcpy( data, data1, size1 );
    }
    MyRingBuffer_AdvanceReadIndex( rbuf, numRead );
    return numRead;
}

typedef struct A_STREAM {
  PaStream *as;
  pthread_mutex_t data_mutex, in_mutex, out_mutex;
  pthread_cond_t in_cond, out_cond;
  MyRingBuffer in_buf, out_buf;
  char *in_data, *out_data;
  int mode;
  PaDeviceIndex id;
  double time, sample_rate, input_latency, output_latency;
#ifdef TIMER_KLUDGE
  double delta;
#endif
  int size, channels, bps, bpf;
  PaSampleFormat format;
  struct A_STREAM *prev, *next;
} a_stream;

static a_stream *current = NULL;

static int init_buf(MyRingBuffer *buf, char **data, long bufsize)
{
  if (!(*data = malloc(bufsize)))
    return 0;
  memset(*data, 0, bufsize);
  if (MyRingBuffer_Init(buf, bufsize, *data)) {
    free(*data);
    return 0;
  }
  return 1;
}

static void fini_buf(MyRingBuffer *buf, char **data)
{
  free(*data);
  *data = NULL;
}

static unsigned long round_pow2(unsigned long n)
{
  /* round to the next power of 2 */
  long numBits = 0;
  if( ((n-1) & n) == 0) return n; /* Already a power of two. */
  while( n > 0 ) {
    n= n>>1;
    numBits++;
  }
  return (1<<numBits);
}

static int init_a_stream(a_stream *v, int mode, long bufsize)
{
  bufsize = round_pow2(bufsize);
  if (mode & MODE_READ)
    if (!init_buf(&v->in_buf, &v->in_data, bufsize))
      return 0;
  if (mode & MODE_WRITE) {
    if (!init_buf(&v->out_buf, &v->out_data, bufsize)) {
      if (mode & MODE_READ) fini_buf(&v->in_buf, &v->in_data);
      return 0;
    }
#if 0
    {
      /* fake a full write buffer */
      long bytes = MyRingBuffer_GetWriteAvailable(&v->out_buf);
      MyRingBuffer_AdvanceWriteIndex(&v->out_buf, bytes);
    }
#endif
  }
  v->mode = mode;
  v->time = 0.0;
#ifdef TIMER_KLUDGE
  v->delta = 0.0;
#endif
  pthread_mutex_init(&v->data_mutex, NULL);
  if (mode & MODE_READ) {
    pthread_mutex_init(&v->in_mutex, NULL);
    pthread_cond_init(&v->in_cond, NULL);
  }
  if (v->mode & MODE_WRITE) {
    pthread_mutex_init(&v->out_mutex, NULL);
    pthread_cond_init(&v->out_cond, NULL);
  }
  if (current) current->prev = v;
  v->prev = NULL;
  v->next = current;
  current = v;
  return 1;
}

static void lock_a_stream(void *x)
{
  a_stream *v = (a_stream*)x;
  pthread_mutex_lock(&v->data_mutex);
  if (v->mode & MODE_READ) pthread_mutex_lock(&v->in_mutex);
  if (v->mode & MODE_WRITE) pthread_mutex_lock(&v->out_mutex);
}

static void unlock_a_stream(void *x)
{
  a_stream *v = (a_stream*)x;
  pthread_mutex_unlock(&v->data_mutex);
  if (v->mode & MODE_READ) pthread_mutex_unlock(&v->in_mutex);
  if (v->mode & MODE_WRITE) pthread_mutex_unlock(&v->out_mutex);
}

static void fini_a_stream(a_stream *v, int abort)
{
  if (v->as) {
    if (abort)
      Pa_AbortStream(v->as);
    else
      Pa_StopStream(v->as);
    pthread_cleanup_push(unlock_a_stream, (void*)v);
    lock_a_stream(v);
    Pa_CloseStream(v->as);
    v->as = NULL;
    /* wake up threads waiting to read or write the stream */
    if (v->mode & MODE_READ) pthread_cond_broadcast(&v->in_cond);
    if (v->mode & MODE_WRITE) pthread_cond_broadcast(&v->out_cond);
    pthread_cleanup_pop(1);
  }
}

static void destroy_a_stream(a_stream *v)
{
  pthread_mutex_destroy(&v->data_mutex);
  if (v->mode & MODE_READ) {
    pthread_mutex_destroy(&v->in_mutex);
    pthread_cond_destroy(&v->in_cond);
  }
  if (v->mode & MODE_WRITE) {
    pthread_mutex_destroy(&v->out_mutex);
    pthread_cond_destroy(&v->out_cond);
  }
  if (v->mode & MODE_READ) fini_buf(&v->in_buf, &v->in_data);
  if (v->mode & MODE_WRITE) fini_buf(&v->out_buf, &v->out_data);
  if (v->prev) v->prev->next = v->next;
  if (v->next) v->next->prev = v->prev;
  if (v == current) current = v->next;
}

FUNCTION(audio,start_audio,argc,argv)
{
  if (argc == 0) {
    a_stream *v;
    for (v = current; v; v = v->next) fini_a_stream(v, 1);
    Pa_Terminate();
    init_ok = (Pa_Initialize() == paNoError);
    if (init_ok)
      return mkvoid;
    else
      return __FAIL;
  } else
    return __FAIL;
}

FUNCTION(audio,stop_audio,argc,argv)
{
  if (argc == 0) {
    a_stream *v;
    for (v = current; v; v = v->next) fini_a_stream(v, 1);
    Pa_Terminate();
    init_ok = 0;
    return mkvoid;
  } else
    return __FAIL;
}

DESTRUCTOR(audio,AudioStream,v)
{
  fini_a_stream((a_stream*)v, 0);
  destroy_a_stream((a_stream*)v);
  free(v);
}

/* Audio processing callback. Note that in difference to the pablio
   implementation we employ some mutex locks and condition variables here.
   This isn't nice but is needed to protect shared data and to wake up a
   thread waiting for data to be read or written. */

static int audio_cb(const void *input, void *output,
		    unsigned long nframes,
		    const PaStreamCallbackTimeInfo *time_info,
		    PaStreamCallbackFlags status,
		    void *data)
{
  a_stream *v = (a_stream*)data;
  long bytes = v->bpf*nframes;
  /* update the current time */
  pthread_mutex_lock(&v->data_mutex);
  if (!v->as) {
    pthread_mutex_unlock(&v->data_mutex);
    return 0;
  }
#ifdef TIMER_KLUDGE
  v->time += v->delta;
  v->delta = ((double)nframes)/v->sample_rate;
#else
  v->time = time_info?time_info->currentTime:0;
#endif
  pthread_mutex_unlock(&v->data_mutex);
  /* process data */
  if (input) {
    pthread_mutex_lock(&v->in_mutex);
    if (MyRingBuffer_GetWriteAvailable(&v->in_buf) == 0)
      MyRingBuffer_AdvanceReadIndex(&v->in_buf, bytes);
    MyRingBuffer_Write(&v->in_buf, input, bytes);
    pthread_cond_signal(&v->in_cond);
    pthread_mutex_unlock(&v->in_mutex);
  }
  if (output) {
    int i, read;
    pthread_mutex_lock(&v->out_mutex);
    read = MyRingBuffer_Read(&v->out_buf, output, bytes);
    if (read < bytes) memset(((char*)output)+read, 0, bytes-read);
    pthread_cond_signal(&v->out_cond);
    pthread_mutex_unlock(&v->out_mutex);
  }
  return 0;
}

FUNCTION(audio,open_audio_stream,argc,argv)
{
  long id, mode, channels, format, size, bps;
  double rate;
  int n;
  expr *xv;
  if (init_ok && argc == 3 &&
      isint(argv[0], &id) && id >= 0 && id <= Pa_GetDeviceCount() &&
      isint(argv[1], &mode) &&
      (mode & ~MODE_RDWR) == 0 && (mode & MODE_RDWR) != 0 &&
      istuple(argv[2], &n, &xv) && n == 4 &&
      (isfloat(xv[0], &rate) || ismpz_float(xv[0], &rate)) && rate > 0.0 &&
      isint(xv[1], &channels) && channels > 0 &&
      isint(xv[2], &format) && (bps = Pa_GetSampleSize(format)) > 0 &&
      isint(xv[3], &size) && size > 0) {
    PaStreamParameters inparams, outparams, *inptr = NULL, *outptr = NULL;
    a_stream *v = malloc(sizeof(a_stream));
    if (!v)
      return __ERROR;
    else if (!init_a_stream(v, mode, bps*channels*size)) {
      free(v);
      return __ERROR;
    }
    if (mode&MODE_READ) {
      inparams.device = (PaDeviceIndex)id;
      inparams.channelCount = channels;
      inparams.sampleFormat = (PaSampleFormat)format;
      inparams.suggestedLatency = 0;
      inparams.hostApiSpecificStreamInfo = NULL;
      inptr = &inparams;
    }
    if (mode&MODE_WRITE) {
      outparams.device = (PaDeviceIndex)id;
      outparams.channelCount = channels;
      outparams.sampleFormat = (PaSampleFormat)format;
      outparams.suggestedLatency = 0;
      outparams.hostApiSpecificStreamInfo = NULL;
      outptr = &outparams;
    }
    if (Pa_OpenStream(&v->as, inptr, outptr,
		      rate, size, paNoFlag, audio_cb, v) ==
	paNoError) {
      const PaStreamInfo* info = Pa_GetStreamInfo(v->as);
#ifdef HAVE_POSIX_SIGNALS
      sigset_t sigset, oldset;
#endif
      v->id = (PaDeviceIndex)id;
      v->sample_rate = info?info->sampleRate:rate;
      v->input_latency = info?info->inputLatency:0;
      v->output_latency = info?info->outputLatency:0;
      v->channels = channels;
      v->format = (PaSampleFormat)format;
      v->size = size;
      v->bps = bps;
      v->bpf = bps*channels;
#ifdef HAVE_POSIX_SIGNALS
      /* temporarily block some signals to make sure that they are blocked
	 in the PortAudio threads created by Pa_StartStream */
      sigemptyset(&sigset);
      sigaddset(&sigset, SIGINT);
      sigaddset(&sigset, SIGQUIT);
      sigaddset(&sigset, SIGTSTP);
      sigaddset(&sigset, SIGTERM);
      sigaddset(&sigset, SIGHUP);
      sigprocmask(SIG_BLOCK, &sigset, &oldset);
#endif
      Pa_StartStream(v->as);
#ifdef HAVE_POSIX_SIGNALS
      sigprocmask(SIG_SETMASK, &oldset, NULL);
#endif
      return mkobj(type(AudioStream), v);
    } else {
      destroy_a_stream(v);
      free(v);
      return __FAIL;
    }
  } else
    return __FAIL;
}

FUNCTION(audio,close_audio_stream,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as) {
    fini_a_stream(v, 0);
    return mkvoid;
  } else
    return __FAIL;
}

FUNCTION(audio,audio_stream_id,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as)
    return mkint((long)v->id);
  else
    return __FAIL;
}

FUNCTION(audio,audio_stream_info,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as)
    return mktuplel(4, mkfloat(v->sample_rate),
		    mkint((long)v->channels),
		    mkint((long)v->format),
		    mkint((long)v->size));
  else
    return __FAIL;
}

FUNCTION(audio,audio_stream_latencies,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as)
    return mktuplel(2, mkfloat(v->input_latency),
		    mkfloat(v->output_latency));
  else
    return __FAIL;
}

FUNCTION(audio,audio_stream_time,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as) {
    double time;
    pthread_mutex_lock(&v->data_mutex);
    time = v->time;
    pthread_mutex_unlock(&v->data_mutex);
    return mkfloat(time);
  } else
    return __FAIL;
}

FUNCTION(audio,bytes_per_sample,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as)
    return mkint(v->bps);
  else
    return __FAIL;
}

FUNCTION(audio,bytes_per_frame,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as)
    return mkint(v->bpf);
  else
    return __FAIL;
}

/* ByteStr data structure, see clib.c */

typedef struct bstr {
  long size;
  unsigned char *v;
} bstr_t;

FUNCTION(audio,read_audio_stream,argc,argv)
{
  a_stream *v;
  long size;
  if (argc == 2 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as && (v->mode&MODE_READ) && isint(argv[1], &size) && size >= 0) {
    bstr_t *m = malloc(sizeof(bstr_t));
    if (!m) return __FAIL;
    if (size > 0) {
      long bytes = size*v->bpf, read;
      unsigned char *p;
      if (!(m->v = malloc(bytes))) {
	free(m);
	return __ERROR;
      }
      m->size = bytes;
      p = m->v;
      release_lock();
      pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock,
			   (void*)&v->in_mutex);
      pthread_mutex_lock(&v->in_mutex);
      brkflag = 0;
      while (!brkflag && v->as && bytes > 0) {
        while (!brkflag && v->as &&
	       (read = MyRingBuffer_Read(&v->in_buf, p, bytes)) == 0)
	  pthread_cond_wait(&v->in_cond, &v->in_mutex);
        bytes -= read;
        p += read;
      }
      pthread_cleanup_pop(1);
      acquire_lock();
      if (brkflag) {
	brkflag = 0;
	free(m->v); free(m);
	return __FAIL;
      }
    } else {
      m->size = 0;
      m->v = NULL;
    }
    return mkobj(type(ByteStr), m);
  } else
    return __FAIL;
}

FUNCTION(audio,write_audio_stream,argc,argv)
{
  a_stream *v;
  bstr_t *m;
  if (argc == 2 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as && (v->mode&MODE_WRITE) &&
      isobj(argv[1], type(ByteStr), (void**)&m)) {
    if (m->size > 0) {
      long size = m->size/v->bpf, bytes = size*v->bpf, written;
      unsigned char *p;
      p = m->v;
      release_lock();
      pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock,
			   (void*)&v->out_mutex);
      pthread_mutex_lock(&v->out_mutex);
      brkflag = 0;
      while (!brkflag && v->as && bytes > 0) {
        while (!brkflag && v->as &&
	       (written = MyRingBuffer_Write(&v->out_buf, p, bytes)) == 0)
	  pthread_cond_wait(&v->out_cond, &v->out_mutex);
        bytes -= written;
        p += written;
      }
      pthread_cleanup_pop(1);
      acquire_lock();
      if (brkflag) {
	brkflag = 0;
	return __FAIL;
      }
    }
    return mkvoid;
  } else
    return __FAIL;
}

FUNCTION(audio,audio_stream_readable,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as && (v->mode & MODE_READ)) {
    int bytes = MyRingBuffer_GetReadAvailable(&v->in_buf);
    return mkint((long)(bytes/v->bpf));
  } else
    return __FAIL;
}

FUNCTION(audio,audio_stream_writeable,argc,argv)
{
  a_stream *v;
  if (argc == 1 && isobj(argv[0], type(AudioStream), (void**)&v) &&
      v->as && (v->mode & MODE_WRITE)) {
    int bytes = MyRingBuffer_GetWriteAvailable(&v->out_buf);
    return mkint((long)(bytes/v->bpf));
  } else
    return __FAIL;
}

INIT(audio)
{
  init_ok = (Pa_Initialize() == paNoError);
  /* register signal handlers */
  int_handler = syssignal(SIGINT, break_handler);
  term_handler = syssignal(SIGTERM, break_handler);
#ifdef SIGHUP
  hup_handler = syssignal(SIGHUP, break_handler);
#endif
}

FINI(audio)
{
  if (init_ok) {
    a_stream *v;
    for (v = current; v; v = v->next) fini_a_stream(v, 1);
    Pa_Terminate();
  }
}
