
/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#if defined (HAVE_CONFIG_H)
#  include "config.h"
#endif

/* system headers */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <time.h>

/* check for standard C headers */
#if STDC_HEADERS
# include <stdlib.h>
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <pthread.h>

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <samplerate.h>
#include <fftw3.h>
#include <portaudio.h>
#include "ggilib.h"
#include <libq.h>

MODULE(wave)

#undef min
#undef max

static inline double min(double x, double y)
{
  if (x <= y)
    return x;
  else
    return y;
}

static inline double max(double x, double y)
{
  if (x >= y)
    return x;
  else
    return y;
}

static inline double sqr(double x)
{
  return x*x;
}

static inline double myabs(double x, double y)
{
  return sqrt(sqr(x)+sqr(y));
}

static inline double myarg(double x, double y)
{
  return atan2(y, x);
}


/* Mersenne Twister RNG by Makoto Matsumoto and Takuji Nishimura, optimized by
   Shawn Cokus. The original sources can be obtained from the Mersenne Twister
   homepage, http://www.math.keio.ac.jp/~matumoto/emt.html */

typedef unsigned long uint32MT;

#define N              (624)
#define M              (397)
#define K              (0x9908B0DFU)
#define hiBit(u)       ((u) & 0x80000000U)
#define loBit(u)       ((u) & 0x00000001U)
#define loBits(u)      ((u) & 0x7FFFFFFFU)
#define mixBits(u, v)  (hiBit(u)|loBits(v))

static uint32MT   state[N+1];
static uint32MT   *next;
static int        left = -1;


static void seedMT(uint32MT seed)
 {
    register uint32MT x = (seed | 1U) & 0xFFFFFFFFU, *s = state;
    register int    j;

    for(left=0, *s++=x, j=N; --j;
        *s++ = (x*=69069U) & 0xFFFFFFFFU);
 }


static uint32MT reloadMT(void)
 {
    register uint32MT *p0=state, *p2=state+2, *pM=state+M, s0, s1;
    register int    j;

    if(left < -1)
        seedMT(4357U);

    left=N-1, next=state+1;

    for(s0=state[0], s1=state[1], j=N-M+1; --j; s0=s1, s1=*p2++)
        *p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);

    for(pM=state, j=M; --j; s0=s1, s1=*p2++)
        *p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);

    s1=state[0], *p0 = *pM ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);
    s1 ^= (s1 >> 11);
    s1 ^= (s1 <<  7) & 0x9D2C5680U;
    s1 ^= (s1 << 15) & 0xEFC60000U;
    return(s1 ^ (s1 >> 18));
 }


static uint32MT randomMT(void)
 {
    uint32MT y;

    if(--left < 0)
        return(reloadMT());

    y  = *next++;
    y ^= (y >> 11);
    y ^= (y <<  7) & 0x9D2C5680U;
    y ^= (y << 15) & 0xEFC60000U;
    return(y ^ (y >> 18));
 }


static double pi, log_2;

INIT(wave)
{
  pi = 4.0*atan(1.0);
  log_2 = log(2.0);
  /* initialize random seed: */
  seedMT(((unsigned long)time(NULL)) << 1 | 1);
}

/* ByteStr data structure, see clib.c */

typedef struct bstr {
  long size;
  unsigned char *v;
} bstr_t;

/* Check for wave objects. */

static int iswave(expr x, long *size, double **data)
{
  bstr_t *m;
  if (isobj(x, type(ByteStr), (void**)&m) && m->size%sizeof(double) == 0) {
    *size = m->size/sizeof(double);
    *data = (double*)m->v;
    return 1;
  } else
    return 0;
}

/* Create a wave object. */

static expr mkwave(long size, double *data)
{
  bstr_t *m = malloc(sizeof(bstr_t));
  if (m) {
    m->size = size*sizeof(double);
    m->v = (unsigned char*)data;
    return mkobj(type(ByteStr), m);
  } else
    return __ERROR;
}

FUNCTION(wave,wave_min,argc,argv)
{
  long size;
  double *wave;
  if (argc == 1 && iswave(argv[0], &size, &wave))
    if (size == 0)
      return mkfloat(0.0);
    else {
      double min = wave[0];
      long i;
      for (i = 1; i < size; i++)
	if (wave[i] < min)
	  min = wave[i];
      return mkfloat(min);
    }
  else
    return __FAIL;
}

FUNCTION(wave,wave_max,argc,argv)
{
  long size;
  double *wave;
  if (argc == 1 && iswave(argv[0], &size, &wave))
    if (size == 0)
      return mkfloat(0.0);
    else {
      double max = wave[0];
      long i;
      for (i = 1; i < size; i++)
	if (wave[i] > max)
	  max = wave[i];
      return mkfloat(max);
    }
  else
    return __FAIL;
}

FUNCTION(wave,wave_rms,argc,argv)
{
  long size;
  double *wave;
  if (argc == 1 && iswave(argv[0], &size, &wave))
    if (size == 0)
      return mkfloat(0.0);
    else {
      double sum = 0.0;
      long i;
      for (i = 0; i < size; i++)
	sum += sqr(wave[i]);
      return mkfloat(sqrt(sum/size));
    }
  else
    return __FAIL;
}

FUNCTION(wave,wave,argc,argv)
{
  if (argc == 1) {
    long n;
    double *wave, val;
    expr x, hd, tl;
    for (n = 0, x = argv[0]; iscons(x, &hd, &tl) &&
	   (isfloat(hd, &val) || ismpz_float(hd, &val)); x = tl)
      n++;
    if (!isnil(x))
      return __FAIL;
    else if (n == 0)
      return mkwave(0, NULL);
    wave = malloc(n*sizeof(double));
    if (!wave)
      return __ERROR;
    for (n = 0, x = argv[0]; iscons(x, &hd, &tl) &&
	   (isfloat(hd, &val) || ismpz_float(hd, &val)); x = tl)
      wave[n++] = val;
    return mkwave(n, wave);
  } else
    return __FAIL;
}

FUNCTION(wave,samples,argc,argv)
{
  long n;
  double *wave;
  if (argc == 1 && iswave(argv[0], &n, &wave)) {
    expr x = mknil;
    while (x && --n >= 0)
      x = mkcons(mkfloat(wave[n]), x);
    return x;
  } else
    return __FAIL;
}

static inline void pack24(void *dest, long val)
{
  unsigned char *mem = (unsigned char*)dest;
  unsigned long temp = val;
#ifdef WORDS_BIGENDIAN
  mem[0] = (unsigned char)(temp >> 24);
  mem[1] = (unsigned char)(temp >> 16);
  mem[2] = (unsigned char)temp >> 8;
#else
  mem[0] = (unsigned char)temp >> 8;
  mem[1] = (unsigned char)(temp >> 16);
  mem[2] = (unsigned char)(temp >> 24);
#endif
}

static inline long unpack24(void *src)
{
  unsigned char *mem = (unsigned char*)src;
  unsigned long temp;
#ifdef WORDS_BIGENDIAN
  temp = mem[2] << 8;
  temp |= mem[1] << 16;
  temp |= mem[0] << 24;
#else
  temp = mem[0] << 8;
  temp |= mem[1] << 16;
  temp |= mem[2] << 24;
#endif
  return temp;
}

FUNCTION(wave,wave_from,argc,argv)
{
  long format;
  bstr_t *m;
  if (argc == 2 && isint(argv[0], &format) &&
      (format == paFloat32 || format == paInt32 || format == paInt24 ||
       format == paInt16 || format == paInt8 || format == paUInt8) &&
      isobj(argv[1], type(ByteStr), (void**)&m)) {
    if (m->size == 0) return mkwave(0, NULL);
    switch (format) {
    case paFloat32: {
      double *dest;
      long i, n = m->size/sizeof(float);
      float *src = (float*)m->v;
      if (m->size%sizeof(float) != 0) return __FAIL;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = src[i];
      return mkwave(n, dest);
    }
    case paInt32: {
      double *dest, fact = 1.0/0x80000000;
      long i, n = m->size/sizeof(long);
      long *src = (long*)m->v;
      if (m->size%sizeof(long) != 0) return __FAIL;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      return mkwave(n, dest);
    }
    case paInt24: {
      double *dest, fact = 1.0/0x80000000;
      long i, n = m->size/3;
      unsigned char *src = m->v;
      if (m->size%3 != 0) return __FAIL;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = fact*unpack24(src+i*3);
      return mkwave(n, dest);
    }
    case paInt16: {
      double *dest, fact = 1.0/0x8000;
      long i, n = m->size/sizeof(short);
      short *src = (short*)m->v;
      if (m->size%sizeof(short) != 0) return __FAIL;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      return mkwave(n, dest);
    }
    case paInt8: {
      double *dest, fact = 1.0/0x80;
      long i, n = m->size;
      signed char *src = (signed char*)m->v;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      return mkwave(n, dest);
    }
    case paUInt8: {
      double *dest, fact = 2.0/0xff;
      long i, n = m->size;
      unsigned char *src = (unsigned char*)m->v;
      dest = malloc(n*sizeof(double));
      if (!dest) return __ERROR;
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i]-1.0;
      return mkwave(n, dest);
    }
    }
  }
}

FUNCTION(wave,wave_to,argc,argv)
{
  long i, n, format;
  double *src;
  if (argc == 2 && isint(argv[0], &format) &&
      (format == paFloat32 || format == paInt32 || format == paInt24 ||
       format == paInt16 || format == paInt8 || format == paUInt8) &&
      iswave(argv[1], &n, &src)) {
    bstr_t *m = malloc(sizeof(bstr_t));
    if (!m) return __ERROR;
    if (n == 0) {
      m->size = 0;
      m->v = NULL;
      return mkobj(type(ByteStr), m);
    }
    switch (format) {
    case paFloat32: {
      float *dest = malloc(n*sizeof(float));
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	dest[i] = src[i];
      m->size = n*sizeof(float);
      m->v = (unsigned char*)dest;
      break;
    }
    case paInt32: {
      double fact = (double)0x7fffffff;
      long *dest = malloc(n*sizeof(long));
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      m->size = n*sizeof(long);
      m->v = (unsigned char*)dest;
      break;
    }
    case paInt24: {
      double fact = (double)0x7fffffff;
      unsigned char *dest = malloc(n*3);
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	pack24(dest+i*3, fact*src[i]);
      m->size = n*3;
      m->v = dest;
      break;
    }
    case paInt16: {
      double fact = (double)0x7fff;
      short *dest = malloc(n*sizeof(short));
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      m->size = n*sizeof(short);
      m->v = (unsigned char*)dest;
      break;
    }
    case paInt8: {
      double fact = (double)0x7f;
      signed char *dest = malloc(n);
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	dest[i] = fact*src[i];
      m->size = n;
      m->v = (unsigned char*)dest;
      break;
    }
    case paUInt8: {
      double fact = 0xff/2.0;
      unsigned char *dest = malloc(n);
      if (!dest) { free(m); return __ERROR; }
      for (i = 0; i < n; i++)
	dest[i] = fact*(src[i]+1.0);
      m->size = n;
      m->v = (unsigned char*)dest;
      break;
    }
    }
    return mkobj(type(ByteStr), m);
  } else
    return __FAIL;
}

FUNCTION(wave,split_wave,argc,argv)
{
  long k, n;
  double *wave;
  if (argc == 2 && isint(argv[0], &k) && k >= 0 &&
      iswave(argv[1], &n, &wave))
    if (k == 0)
      if (n == 0)
	return mknil;
      else
	return __FAIL;
    else if (k <= n && n%k == 0) {
      long i, j, m = n/k;
      expr *xv = malloc(k*sizeof(expr));
      if (!xv) return __ERROR;
      for (i = 0; i < k; i++) {
	double *channel = malloc(m*sizeof(double));
	if (!channel) {
	  for (j = 0; j < i; j++)
	    dispose(xv[j]);
	  free(xv);
	  return __ERROR;
	}
	for (j = 0; j < m; j++)
	  channel[j] = wave[k*j+i];
	xv[i] = mkwave(m, channel);
	if (!xv[i]) {
	  for (j = 0; j < i; j++)
	    dispose(xv[j]);
	  free(xv);
	  return __ERROR;
	}
      }
      return mklistv(k, xv);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,join_wave,argc,argv)
{
  expr x, hd, tl;
  int k;
  long i, j, n, m;
  double *wave, *channel;
  if (argc != 1) return __FAIL;
  for (k = 0, x = argv[0];
       iscons(x, &hd, &tl) && iswave(hd, &m, &channel);
       x = tl) {
    if (k == 0)
      n = m;
    else if (n != m)
      return __FAIL;
    k++;
  }
  if (!isnil(x)) return __FAIL;
  if (k == 0) return mkwave(0, NULL);
  n = k*n;
  wave = malloc(n*sizeof(double));
  if (!wave) return __ERROR;
  for (i = 0, x = argv[0];
       iscons(x, &hd, &tl) && iswave(hd, &m, &channel);
       x = tl) {
    for (j = 0; j < m; j++)
      wave[k*j+i] = channel[j];
    i++;
  }
  return mkwave(n, wave);
}

FUNCTION(wave,cst_wave,argc,argv)
{
  double a;
  long i, n;
  if (argc == 2 && (isfloat(argv[0], &a) || ismpz_float(argv[0], &a)) &&
      isint(argv[1], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = a;
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,line_wave,argc,argv)
{
  double a, b;
  long i, n;
  if (argc == 3 && (isfloat(argv[0], &a) || ismpz_float(argv[0], &a)) &&
      (isfloat(argv[1], &b) || ismpz_float(argv[1], &b)) &&
      isint(argv[2], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double d = (b-a)/(n-1), *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = a+i*d;
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,sine_wave,argc,argv)
{
  long i, n;
  double k;
  if (argc == 2 && isint(argv[0], &n) && n >= 0 &&
      (isfloat(argv[1], &k) || ismpz_float(argv[1], &k)) && k > 0.0)
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else {
      double f = 1.0/k, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = sin(2*pi*f*i);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,cosine_wave,argc,argv)
{
  long i, n;
  double k;
  if (argc == 2 && isint(argv[0], &n) && n >= 0 &&
      (isfloat(argv[1], &k) || ismpz_float(argv[1], &k)) && k > 0.0)
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else {
      double f = 1.0/k, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = cos(2*pi*f*i);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

static double saw(double f, double x)
{
  double ip, fp = modf(f*x, &ip);
  return -1.0+2*fp;
}

FUNCTION(wave,saw_wave,argc,argv)
{
  long i, n;
  double k;
  if (argc == 2 && isint(argv[0], &n) && n >= 0 &&
      (isfloat(argv[1], &k) || ismpz_float(argv[1], &k)) && k > 0.0)
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else {
      double f = 1.0/k, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = saw(f, (double)i);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

static double triangle(double f, double x)
{
  double ip, fp = modf(f*x, &ip);
  if (fp <= 0.5)
    return 4*fp-1.0;
  else
    return 1.0-4*(fp-0.5);
}

FUNCTION(wave,triangle_wave,argc,argv)
{
  long i, n;
  double k;
  if (argc == 2 && isint(argv[0], &n) && n >= 0 &&
      (isfloat(argv[1], &k) || ismpz_float(argv[1], &k)) && k > 0.0)
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else {
      double f = 1.0/k, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = triangle(f, (double)i);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

static double square(double f, double g, double x)
{
  double ip, fp = modf(f*x, &ip);
  if (fp < g)
    return 1.0;
  else
    return -1.0;
}

FUNCTION(wave,square_wave,argc,argv)
{
  long i, n;
  double k, l;
  if (argc == 3 && isint(argv[0], &n) && n >= 0 &&
      (isfloat(argv[1], &k) || ismpz_float(argv[1], &k)) && k > 0.0 &&
      (isfloat(argv[2], &l) || ismpz_float(argv[2], &l)))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else {
      double f = 1.0/k, g = l/k, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = square(f, g, (double)i);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,random_wave,argc,argv)
{
  long i, n;
  if (argc == 1 && isint(argv[0], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++) {
	double randval = ((double)randomMT())/4294967295.0;
	randval += ((double)randomMT())/4294967295.0/4294967296.0;
	wave[i] = 2.0*randval-1.0;
      }
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,welch_window,argc,argv)
{
  long i, n;
  if (argc == 1 && isint(argv[0], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double n2 = n/2.0, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++) {
	double x = (i-n2)/n2;
	wave[i] = 1.0-x*x;
      }
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,bartlett_window,argc,argv)
{
  long i, n;
  if (argc == 1 && isint(argv[0], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double n2 = n/2.0, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	if (i <= n2)
	  wave[i] = 1.0+2.0*(i-n2)/n;
	else
	  wave[i] = 1.0-2.0*(i-n2)/n;
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,hanning_window,argc,argv)
{
  long i, n;
  if (argc == 1 && isint(argv[0], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double n2 = n/2.0, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = 0.5+0.5*cos(2*pi*(i-n2)/n);
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,hamming_window,argc,argv)
{
  long i, n;
  if (argc == 1 && isint(argv[0], &n))
    if (n == 0)
      return mkwave(0, NULL);
    else if (n > (long)(INT_MAX/sizeof(double)))
      return __ERROR;
    else if (n > 0) {
      double n2 = n/2.0, *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = 0.54+0.46*cos(2.0*pi*(i-n2)/n);
      return mkwave(n, wave);
    } else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,reverse_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = arg[n-i-1];
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,neg_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = -arg[i];
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,inv_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = 1.0/arg[i];
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,abs_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = (arg[i]>=0)?arg[i]:-arg[i];
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,sqrt_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = sqrt(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,exp_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = exp(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,ln_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = log(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,log_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = log10(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,lg_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = log(arg[i])/log_2;
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,sin_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = sin(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,cos_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = cos(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,tan_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = tan(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,asin_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = asin(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,acos_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = acos(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,atan_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = atan(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,sinh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = sinh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,cosh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = cosh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,tanh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = tanh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

#ifndef HAVE_ASINH
static inline double asinh(double x)
{
  return log(x+sqrt(x*x+1));
}
#endif

FUNCTION(wave,asinh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = asinh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

#ifndef HAVE_ACOSH
static inline double acosh(double x)
{
  return log(x+sqrt(x*x-1));
}
#endif

FUNCTION(wave,acosh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = acosh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

#ifndef HAVE_ATANH
static inline double atanh(double x)
{
  return log((1.0+x)/(1.0-x))/2.0;
}
#endif

FUNCTION(wave,atanh_wave,argc,argv)
{
  long i, n;
  double *arg;
  if (argc == 1 && iswave(argv[0], &n, &arg))
    if (n == 0)
      return mkwave(0, NULL);
    else {
      double *wave = malloc(n*sizeof(double));
      if (!wave) return __ERROR;
      for (i = 0; i < n; i++)
	wave[i] = atanh(arg[i]);
      return mkwave(n, wave);
    }
  else
    return __FAIL;
}

FUNCTION(wave,sum_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]+arg2[i];
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = a1+arg2[i];
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]+a2;
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,diff_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]-arg2[i];
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = a1-arg2[i];
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]-a2;
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,prd_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]*arg2[i];
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = a1*arg2[i];
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]*a2;
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,quot_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]/arg2[i];
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = a1/arg2[i];
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = arg1[i]/a2;
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,pow_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = exp(log(arg1[i])*arg2[i]);
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = exp(log(a1)*arg2[i]);
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = exp(log(arg1[i])*a2);
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,min_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = min(arg1[i], arg2[i]);
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = min(a1, arg2[i]);
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = min(arg1[i], a2);
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

FUNCTION(wave,max_wave,argc,argv)
{
  long i, n, m;
  double a1, a2, *arg1, *arg2;
  if (argc == 2)
    if (iswave(argv[0], &n, &arg1) && iswave(argv[1], &m, &arg2) &&
	n == m)
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = max(arg1[i], arg2[i]);
	return mkwave(n, wave);
      }
    else if ((isfloat(argv[0], &a1) || ismpz_float(argv[0], &a1)) &&
	     iswave(argv[1], &n, &arg2))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = max(a1, arg2[i]);
	return mkwave(n, wave);
      }
    else if (iswave(argv[0], &n, &arg1) && 
	     (isfloat(argv[1], &a2) || ismpz_float(argv[1], &a2)))
      if (n == 0)
	return mkwave(0, NULL);
      else {
	double *wave = malloc(n*sizeof(double));
	if (!wave) return __ERROR;
	for (i = 0; i < n; i++)
	  wave[i] = max(arg1[i], a2);
	return mkwave(n, wave);
      }
    else
      return __FAIL;
  else
    return __FAIL;
}

typedef struct {
  pthread_mutex_t mutex;
  SRC_STATE* state;
  SRC_DATA data;
  double last_ratio;
  long channels, input_size, output_size;
} src_state_t;

DESTRUCTOR(wave,SRCState,x)
{
  src_state_t *state = (src_state_t*)x;
  pthread_mutex_destroy(&state->mutex);
  src_delete(state->state);
  if (state->data.data_in)
    free(state->data.data_in);
  if (state->data.data_out)
    free(state->data.data_out);
  free(state);
}

FUNCTION(wave,src_new,argc,argv)
{
  long conv_type, channels;
  double ratio;
  if (argc == 2 &&
      isint(argv[0], &conv_type) && conv_type >= 0 && conv_type <= 4 &&
      isint(argv[1], &channels) && channels > 0) {
    int err;
    src_state_t *state = malloc(sizeof(src_state_t));
    if (!state) return __ERROR;
    state->state = src_new(conv_type, channels, &err);
    if (!state->state) {
      free(state);
      return __FAIL;
    }
    pthread_mutex_init(&state->mutex, NULL);
    state->channels = channels;
    state->input_size = state->output_size = 0;
    state->last_ratio = 0.0;
    state->data.data_in = state->data.data_out = NULL;
    state->data.input_frames = state->data.output_frames = 0;
    state->data.input_frames_used = state->data.output_frames_gen = 0;
    state->data.end_of_input = 0;
    state->data.src_ratio = 0.0;
    return mkobj(type(SRCState), state);
  } else
    return __FAIL;
}

FUNCTION(wave,src_process,argc,argv)
{
  src_state_t *state;
  double ratio, *wave;
  long i, n, m, p;
  if (argc == 3 && isobj(argv[0], type(SRCState), (void**)&state) &&
      (isfloat(argv[1], &ratio) || ismpz_float(argv[1], &ratio)) &&
      ratio > 0.0 && src_is_valid_ratio(ratio) &&
      iswave(argv[2], &n, &wave) && n > 0 && n%state->channels == 0) {
    expr ret = NULL;
    int res, done = 0;
    float *input, *output, *inptr, *outptr;
    /* Release the lock here, since this operation may take some time. We then
       acquire the mutex of the SRCState object, so that no other thread
       messes with the internal state as long as the conversion is in
       progress. */
    release_lock();
    pthread_cleanup_push((void(*)(void*))pthread_mutex_unlock,
			 (void*)&state->mutex);
    pthread_mutex_lock(&state->mutex);
    /* update the conversion ratio */
    state->data.src_ratio = ratio;
    if (n != state->input_size) {
      /* reallocate the input buffer */
      float *x;
      x = realloc(state->data.data_in, n*sizeof(float));
      if (!x) { ret = __ERROR; goto error; }
      state->data.data_in = x;
      state->data.input_frames = n/state->channels;
      state->input_size = n;
    }
    /* Guestimate an output buffer size. Note that we always take the maximum
       of the sizes for the current and the previous conversion ratio, to
       account for libsamplerate interpolating the sample rate if the
       conversion ratio varies. */
    m = (long)(((ratio>state->last_ratio)?ratio:state->last_ratio)*n);
    /* align to the next frame boundary */
    p = m%state->channels;
    if (p != 0)
      m += state->channels-p;
    /* As a cautionary measure, add a couple of extra frames, to account for
       potential rounding discrepancies and wrapup at the beginning/end of the
       conversion. By this we hopefully prevent additional reallocations of
       the output buffer later, at least in most cases. */
    m += state->channels*1024;
    if (m != state->output_size) {
      /* reallocate the output buffer */
      float *y;
      y = realloc(state->data.data_out, m*sizeof(float));
      if (!y) { ret = __ERROR; goto error; }
      state->data.data_out = y;
      state->data.output_frames = m/state->channels;
      state->output_size = m;
    }
    /* keep track of the previous conversion ratio */
    state->last_ratio = ratio;
    /* transfer the input data */
    for (i = 0; i < n; i++)
      state->data.data_in[i] = (float)wave[i];
    /* The actual conversion starts here. The I/O buffers are saved in the
       input and output variables (we'll have to restore these values later),
       while inptr and outptr always point to the current input/output
       position. */
    inptr = input = state->data.data_in;
    outptr = output = state->data.data_out;
#ifdef DEBUG
    printf("-> starting conversion (total: %d frames, output buffer: %d frames),\n\
   input=%p (%d frames read), output=%p (%d frames written)\n",
	   n/state->channels, m/state->channels,
	   input, (inptr-input)/state->channels,
	   output, (outptr-output)/state->channels);
#endif
    res = src_process(state->state, &state->data);
    while (!res) {
      float *y;
      /* update input/output pointers */
      inptr += state->data.input_frames_used*state->channels;
      outptr += state->data.output_frames_gen*state->channels;
#ifdef DEBUG
      printf("-> %d/%d input frames processed, %d/%d output frames generated,\n\
   input=%p (%d frames read), output=%p (%d frames written)\n",
	     state->data.input_frames_used,
	     state->data.input_frames,
	     state->data.output_frames_gen,
	     state->data.output_frames,
	     input, (inptr-input)/state->channels,
	     output, (outptr-output)/state->channels);
#endif
      /* check whether the entire input has been processed */
      if (state->data.input_frames_used >= state->data.input_frames)
	break;
      /* there are additional input samples to be processed; enlarge the
	 output buffer */
      m += state->channels*1024;
      y = realloc(output, m*sizeof(float));
      if (!y) {
	/* restore I/O buffers and bail out */
	state->data.data_in = input;
	state->data.data_out = output;
	state->data.input_frames = n/state->channels;
	state->data.output_frames = m/state->channels;
	ret = __ERROR;
	goto error;
      }
      state->output_size = m;
      /* set up for the next call */
      state->data.data_in = inptr;
      state->data.input_frames -= state->data.input_frames_used;
      outptr = y+(outptr-output);
      output = y;
      state->data.output_frames = output+m/state->channels-outptr;
      state->data.data_out = outptr;
#ifdef DEBUG
      printf("-> new output buffer %p (%d frames)\n", output, m/state->channels);
#endif
      /* process remaining samples */
      res = src_process(state->state, &state->data);
    }
    /* Calculate the total amount of converted samples and restore the I/O
       buffers. */
    state->data.output_frames_gen = (outptr-output)/state->channels;
#ifdef DEBUG
    printf("-> conversion finished, a total of %d frames were generated\n",
	   state->data.output_frames_gen);
#endif
    state->data.data_in = input;
    state->data.data_out = output;
    state->data.input_frames = n/state->channels;
    state->data.output_frames = m/state->channels;
    /* handle end of input */
    if (state->data.end_of_input) {
      src_reset(state->state);
      state->data.end_of_input = 0;
    }
    /* check for conversion failures */
    if (res) goto error;
    /* get hold of the output data */
    m = state->channels*state->data.output_frames_gen;
    done = 1;
    if (m == 0)
      wave = NULL;
    else {
      wave = malloc(m*sizeof(double));
      if (!wave) { done = 0; ret = __ERROR; goto error; }
      for (i = 0; i < m; i++)
	wave[i] = (double)state->data.data_out[i];
    }
  error: ;
    /* We're done. (This point is also reached if there was a memory
       allocation error.) Acquire the lock again and return the result. */
    pthread_cleanup_pop(1);
    acquire_lock();
    if (done)
      return mkwave(m, wave);
    else
      return ret;
  } else
    return __FAIL;
}

FUNCTION(wave,src_process1,argc,argv)
{
  src_state_t *state;
  double ratio, *wave;
  long n;
  if (argc == 3 && isobj(argv[0], type(SRCState), (void**)&state) &&
      (isfloat(argv[1], &ratio) || ismpz_float(argv[1], &ratio)) &&
      ratio > 0.0 && src_is_valid_ratio(ratio) &&
      iswave(argv[2], &n, &wave) && n > 0 && n%state->channels == 0) {
    pthread_mutex_lock(&state->mutex);
    state->data.end_of_input = 1;
    pthread_mutex_unlock(&state->mutex);
    return FUNCALL(wave, src_process, argc, argv);
  } else
    return __FAIL;
}

FUNCTION(wave,src_set_ratio,argc,argv)
{
  src_state_t *state;
  double ratio;
  if (argc == 2 && isobj(argv[0], type(SRCState), (void**)&state) &&
      (isfloat(argv[1], &ratio) || ismpz_float(argv[1], &ratio)) &&
      ratio > 0.0 && src_is_valid_ratio(ratio)) {
    expr ret = NULL;
    pthread_mutex_lock(&state->mutex);
    if (ratio == state->last_ratio)
      ret = mkvoid;
    else if (!src_set_ratio(state->state, ratio)) {
      state->last_ratio = ratio;
      ret = mkvoid;
    }
    pthread_mutex_unlock(&state->mutex);
    return ret;
  } else
    return __FAIL;
}

FUNCTION(wave,src_reset,argc,argv)
{
  src_state_t *state;
  if (argc == 1 && isobj(argv[0], type(SRCState), (void**)&state)) {
    int res;
    pthread_mutex_lock(&state->mutex);
    res = src_reset(state->state);
    pthread_mutex_unlock(&state->mutex);
    if (res)
      return __FAIL;
    else
      return mkvoid;
  } else
    return __FAIL;
}

static int fft(int n, double *wave, double *mag, double *phase)
{
  int i, n2 = n/2;
  double n_2 = ((double)n)/2.0;
  double *x;
  fftw_plan p;

  x = fftw_malloc(n*sizeof(double));
  if (!x)
    return 0;
  p = fftw_plan_r2r_1d(n, x, x, FFTW_R2HC, FFTW_ESTIMATE);
  if (!p) {
    fftw_free(x);
    return 0;
  }
  memcpy(x, wave, n*sizeof(double));
  fftw_execute(p);
  mag[0] = myabs(x[0], 0.0)/n;
  phase[0] = myarg(x[0], 0.0);
  for (i = 1; i < n2; i++) {
    mag[i] = myabs(x[i], x[n-i])/n_2;
    phase[i] = myarg(x[i], x[n-i]);
  }
  if (n > 1)
    if (n%2 == 0) {
      mag[n2] = myabs(x[n2], 0.0)/n;
      phase[n2] = myarg(x[n2], 0.0);
    } else {
      mag[n2] = myabs(x[n2], x[n-n2])/n_2;
      phase[n2] = myarg(x[n2], x[n-n2]);
    }
  fftw_destroy_plan(p);
  fftw_free(x);
  return 1;
}

FUNCTION(wave,fft,argc,argv)
{
  double *wave, *mag, *phase;
  long n, m;
  if (argc == 1 && iswave(argv[0], &n, &wave) && n > 0) {
    m = n/2+1;
    mag = malloc(m*sizeof(double));
    phase = malloc(m*sizeof(double));
    if (!mag || !phase || !fft(n, wave, mag, phase)) {
      if (mag) free(mag);
      if (phase) free(phase);
      return __ERROR;
    }
    return mktuplel(2, mkwave(m, mag), mkwave(m, phase));
  } else
    return __FAIL;
}

static int ifft(int n, double *mag, double *phase, double *wave)
{
  int i, n2 = n/2;
  double *x;
  fftw_plan p;

  x = fftw_malloc(n*sizeof(double));
  if (!x)
    return 0;
  p = fftw_plan_r2r_1d(n, x, x, FFTW_HC2R, FFTW_ESTIMATE);
  if (!p) {
    fftw_free(x);
    return 0;
  }
  x[0] = mag[0]*cos(phase[0]);
  for (i = 1; i < n2; i++) {
    x[i] = mag[i]*cos(phase[i])/2.0;
    x[n-i] = mag[i]*sin(phase[i])/2.0;
  }
  x[n2] = mag[n2]*cos(phase[n2]);
  fftw_execute(p);
  memcpy(wave, x, n*sizeof(double));
  fftw_destroy_plan(p);
  fftw_free(x);
  return 1;
}

FUNCTION(wave,ifft,argc,argv)
{
  double *wave, *mag, *phase;
  long n, m;
  expr *xv;
  int k;
  if (argc == 1 && istuple(argv[0], &k, &xv) && k == 2 &&
      iswave(xv[0], &n, &mag) && n > 0 &&
      iswave(xv[1], &m, &phase) && m == n) {
    n = 2*(n-1);
    if (n == 0)
      return mkwave(0, NULL);
    wave = malloc(n*sizeof(double));
    if (!wave)
      return __ERROR;
    else if (!ifft(n, mag, phase, wave)) {
      free(wave);
      return __ERROR;
    }
    return mkwave(n, wave);
  } else
    return __FAIL;
}

typedef struct {
  ggi_visual_t vis;
} visual_t;

static void shade(ggi_color *c)
{
  if (c->r || c->g || c->b) {
    c->r -= c->r>>2;
    c->g -= c->g>>2;
    c->b -= c->b>>2;
  } else {
    c->r = c->g = c->b = 0x4000;
  }
}

FUNCTION(wave,draw_wave,argc,argv)
{
  visual_t *v;
  ggi_visual_t vis;
  expr *xv;
  int m;
  long x, y, w, h, n;
  double *wave;
  if (argc == 4 &&
      isobj(argv[0], type(GGIVisual), (void**)&v) && (vis = v->vis) &&
      istuple(argv[1], &m, &xv) && m == 2 &&
      isint(xv[0], &x) && isint(xv[1], &y) &&
      istuple(argv[2], &m, &xv) && m == 2 &&
      isint(xv[0], &w) && isint(xv[1], &h) &&
      w > 0 && h > 0 &&
      iswave(argv[3], &n, &wave)) {
    if (n <= 0)
      return mkvoid;
    else if (n <= w) {
      double dx = ((double)w-1)/(n-1);
      long i, k;
      int show_dots = (dx>=5.0), xa, ya;
      for (i = k = 0; i < n && k < w; i++, k++) {
	int val = (int)((wave[i]+1.0f)/2.0f*(h-1));
	int xe = (int)x+k*dx, ye = y+h-1-val;
	if (k > 0 && draw_line(v, xa, ya, xe, ye))
	  return __FAIL;
	if (show_dots && draw_box(v, xe-1, ye-1, 3, 3) < 0)
	  return __FAIL;
	xa = xe; ya = ye;
      }
#ifdef DEBUG
      printf("%d frames, 1 sample per pixel\n", n);
#endif
    } else {
      int p = n/w+((n%w==0)?0:1), vals = n/p+((n%p==0)?0:1);
      int *min = malloc(vals*sizeof(int)), *max = malloc(vals*sizeof(int)),
	*rmsh = malloc(vals*sizeof(int)), *rmsl = malloc(vals*sizeof(int));
      double dx = ((double)w-1)/(vals-1), sum, mean;
      long i, j, k, l;
      ggi_pixel pix, pix2;
      ggi_color col;
      if (!min || !max || !rmsh || !rmsl) {
	if (min) free(min);
	if (max) free(max);
	if (rmsh) free(rmsh);
	if (rmsl) free(rmsl);
	return __ERROR;
      }
      for (i = k = 0; k < vals; i += p, k++) {
	min[k] = w; max[k] = -1; sum = 0.0; l = 0;
	for (j = i; j < i+p && j < n; j++) {
	  double samp = wave[j];
	  int val = (int)((samp+1.0f)/2.0f*(h-1));
	  if (val<min[k]) min[k] = val;
	  if (val>max[k]) max[k] = val;
	  sum += samp*samp; l++;
	}
	mean = sqrt(sum/l);
	rmsh[k] = (int)((mean+1.0f)/2.0f*(h-1));
	rmsl[k] = (int)((-mean+1.0f)/2.0f*(h-1));
	if (k > 0) {
	  if (max[k] < min[k-1])
	    max[k] = min[k-1]-1;
	  if (min[k] > max[k-1])
	    min[k] = max[k-1]+1;
	}
	if (min[k] < 0) min[k] = 0;
	if (min[k] >= w) min[k] = h-1;
	if (max[k] < 0) max[k] = 0;
	if (max[k] >= w) max[k] = h-1;
	if (min[k] > max[k]) min[k] = max[k];
	if (rmsh[k] >= max[k]) rmsh[k] = max[k]-1;
	if (rmsl[k] <= min[k]) rmsl[k] = min[k]+1;
      }
      ggiGetGCForeground(vis, &pix);
      if (ggiUnmapPixel(vis, pix, &col)) return __FAIL;
      shade(&col);
      pix2 = ggiMapColor(vis, &col);
      for (k = 0; k < vals; k++) {
	int xa, xe = (int)x+(k+1)*dx;
	for (xa = (int)x+k*dx; xa < xe; xa++)
	  if (draw_vline(v, xa, y+h-1-max[k], max[k]-min[k]+1) < 0)
	    return __FAIL;
      }
      ggiSetGCForeground(vis, pix2);
      for (k = 0; k < vals; k++) {
	int xa, xe = (int)x+(k+1)*dx;
	if (rmsl[k] <= rmsh[k])
	  for (xa = (int)x+k*dx; xa < xe; xa++)
	    if (draw_vline(v, xa, y+h-1-rmsh[k], rmsh[k]-rmsl[k]+1) < 0)
	      return __FAIL;
      }
      ggiSetGCForeground(vis, pix);
      free(min); free(max); free(rmsh); free(rmsl);
#ifdef DEBUG
      printf("%d frames, %d samples per pixel\n", n, p);
#endif
    }
    return mkvoid;
  } else
    return __FAIL;
}
